###########################################################################################
#
# How to add a clustering engine ?
#
##################################
#
# Simply creates a class in a new module that
# implements vizuka.clustering.clusterizer.Clusterizer
#
# Below is a simple example, with kmeans
# If you need special parameters, pass them in vizualization.py (search "make_clusterizer")
# If these special parameters are hyperparameters see "required_arguments" below
#
#
##########################################################################################


###################
#
# WORKING EXAMPLE :
#
###################
#
# from sklearn.cluster import KMeans
# from vizuka.clustering.clusterizer import Clusterizer
#
#
# class KmeansClusterizer(Clusterizer):
#
#     required_arguments = ['Number of clusters']
#     # will be asked in popup dialog when running clustering
# 
#     def __init__(self, required_arguments={'Number of clusters':15}):
#         """
#         Uses sklearn kmeans, accepts same arguments.
#         Default nb of cluster : 120
#         """
#         self.engine = KMeans(
#                 n_clusters=required_arguments['Number of clusters'],
#                 )
#         self.method='kmeans'
# 
#     def fit(self, xs):
#         """
#         Fit the datas and find clusterization adapted to the data provided
#
#         :param xs: data to clusterize
#         """
#         self.engine.fit(xs)
#
#     def predict(self, xs):
#         """
#         Predicts cluster label
#
#         :param xs: array-like of datas
#         :return:   list of cluster possible_outputs_list
#         """
#         return self.engine.predict(xs)



# HOW TO ADD A CLUSTER VIZUALIZATION ?
######################################
# 
# If you want to add another tool toi visualize the inner observations inside a selected cluster
# you can add here an new tool.
# 
# You should inherits vizuka.cluster_viewer.plotter.Plotter, and define a callable,
# cf the simple the example below.
# Three arguments are passed to you : data array , a fig and a gridspec
#
# Use them as follow :
#    - at the beginning : axe = plt.Subplot(fig, spec)
#    - do stuff         : axe.plot(do_stuff(data))
#    - at the end       : fig.add_subplot(axe)
# 
# Call it at vizuka runtime : vizuka -s COLUMN_NAME:this_filename
#
##########################################################################################

#################
# WORKING EXAMPLE
#################
#
# from collections import Counter
# 
# import numpy as np
# import matplotlib
# matplotlib.use('Qt5Agg')  # noqa
# from matplotlib import pyplot as plt
# 
# from vizuka.cluster_viewer.plotter import Plotter
# 
# class Counter(Plotter):
# 
#     def __call__(self, data, fig, spec):
#         """
#         Show a counter of each class occurrences
# 
#         :param: data is a list of the observations
#         :param: fig is the Figure on which the axe will be drawn
#         :param:spec is the GridSpec for the axe
# 
#         :return: the axe with the visualization
#         """
#         if not data:
#             return
#         axe = plt.Subplot(fig, spec)
# 
#         c = Counter(data)
#         x = [l for l in c.keys()]
#         y = [c[l] for l in x]
# 
#         
#         order = np.argsort(y)
#         y = [y[i] for i in order]
#         x = [x[i] for i in order]
# 
#         graduation = np.linspace(0, len(y), len(y))
#         axe.bar(height=y, left=graduation)
#         axe.set_xticks(graduation)
#         axe.set_xticklabels([str(i) for i in x])
#         fig.add_subplot(axe)
# 
#         return axe
#
#######################################################################



"""
Cool functions for ML
"""
import math
import numpy as np

def entropy(my_dic):
    """
    StraightForward entropy calculation

    :param my_dict: dict of occurence of different classes
    :return: discrete entropy calculation
    """
    effectif_total = sum(my_dic.values())
    s = 0
    for effectif in my_dic.values():
        proportion = effectif / effectif_total
        if proportion > 0:
            s += proportion * math.log(proportion)
    return -s

def cross_entropy(global_distribution, specialized_distribution, global_entropy=None):
    """
    Cross-entropy between two dicts
    global_dictionnary must contains all keys in specialized_distribution

    :param gobal_distribution: dictionnary containing {class_label:occurence}
    :param specialized_distribution: dictionnary containing {class_label:occurence}
    """

    specialized_array = np.array(specialized_distribution)

    if global_entropy is None:
        global_array = np.array(global_distribution)
        global_entropy = np.log(global_array / np.sum(global_array))
        if len(global_array) == 0:  # case where there's nothing
            return 0

    entropy_local = specialized_array / np.sum(specialized_array)
    
    return np.sum( - entropy_local * global_entropy)



import random

import matplotlib
matplotlib.use('Qt5Agg')  # noqa
from matplotlib import pyplot as plt
from matplotlib import gridspec

from vizuka.cluster_viewer.plotter import Plotter

class RandomImages(Plotter):

    def __call__(self, data, fig, spec):
        """
        Plots some random images found in the :param data: list given

        :param: data is a list of the observations
        :param: fig is the Figure on which the axe will be drawn
        :param:spec is the GridSpec for the axe

        :return: the axe with the visualization
        """
        if not data:
            return

        selected_data = [data[random.randint(0,max(len(data)-1,0))] for _ in range(25)]
        inner = gridspec.GridSpecFromSubplotSpec(5,5,
                        subplot_spec=spec)
        for idx, inner_spec in enumerate(inner):
            axe = plt.Subplot(fig, inner_spec)
            axe.imshow(selected_data[idx])
            fig.add_subplot(axe)
        return axe



import itertools
import os
import logging

import numpy as np
# from MulticoreTSNE import MulticoreTSNE as tsne_algorithm
from sklearn.manifold import TSNE as tsne_algorithm

from vizuka.config import (
    VERSION,
    REDUCED_DATA_PATH,
    REDUCED_DATA_NAME,
)
from vizuka.dimension_reduction import (
        projector,
        tsne,
        )


class tSNE(projector.Projector):

    def __init__(self, perplexity, learning_rate, n_iter):
        
        self.perplexity     = perplexity
        self.learning_rate  = learning_rate
        self.n_iter         = n_iter

        self.parameters = [perplexity, learning_rate, n_iter]

        self.engine = tsne_algorithm(
            perplexity    = self.perplexity,
            learning_rate = self.learning_rate,
            n_iter        = self.n_iter,
            # n_jobs=3, # only use with Multicore_tSNE !
            )
        self.projections = []

    def project(self, x):
        self.projections = self.engine.fit_transform(x)
        return y



import argparse
import logging

from vizuka import dimension_reduction
from vizuka import data_loader
from vizuka import config

from vizuka.config import (
    VERSION,
    REDUCED_DATA_PATH,
    INPUT_FILE_BASE_NAME,
    DATA_PATH,
    BASE_PATH,
    PROJECTION_DEFAULT_PARAMS,
)

def do_reduce(algorithm_name, parameters, version, data_path, reduced_path):
    """
    Project the data in :param data_path: (version :param version:)
    with algorithm :param algorithm_name: using parameters in :param parameters
    and save it in :param reduced_path:
    """
    algo_builder = dimension_reduction.make_projector(algorithm_name)
    algo = algo_builder(**parameters)


    (x, _ ,_ ,_ ) = data_loader.load_preprocessed(
            file_base_name   = INPUT_FILE_BASE_NAME,
            path             = data_path,
            version          = version,
            )

    algo.project(x)                                   # do the dimension projection
    algo.save(version=version, path=reduced_path)     # save the result

def main():
    """
    Loads parameters and run do_reduce
    """

    logging.basicConfig(level=logging.DEBUG)

    builtin_projectors, extra_projectors = dimension_reduction.list_projectors()

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-a', '--algorithm',
        help=(
            'algorithm name to reduce dimensions, available are :\n'
             + '\t\t\tbuiltin: {}\n'.format(builtin_projectors)
             + '\t\t\textra plugins: {}'.format(extra_projectors)
             )
         )
    parser.add_argument(
        '-p', '--parameters', action='append',
        help='specify parameters, e.g: -p perplexity:50 -p learning_rate:0.1\n'
             '\t\t\twill load default values in config.py if not specified'
        )
    parser.add_argument(
        '-v', '--version',
        help='specify a version of the files to load/generate (see vizuka --show_required_files), currently: '+VERSION)
    parser.add_argument(
        '--path',
        help='change the location of your data/ folder containing reduced/ (the projections)'
             ' and set/ (the raw and preprocessed data)'
             'default to {}'.format(DATA_PATH)
        )

    parser.set_defaults(
            algorithm = 'tsne',
            parameters    = {},
            path      = BASE_PATH,
            version   = VERSION,
            )

    args = parser.parse_args()

    (data_path, reduced_path, _, _, _, _) = config.path_builder(args.path)

    algorithm_name = args.algorithm
    parameters     = args.parameters
    version        = args.version

    for raw_param in parameters:
        if ':' not in raw_name:
            raise TypeError('parameter -p not used correctly! see --help')
        param_name, param_value = raw_param.split(':')
        parameters[param_name:param_value]

    default_params = PROJECTION_DEFAULT_PARAMS.get(algorithm_name, {})
    for default_param_name, default_param_value in default_params.items():
        if default_param_name not in parameters:
            parameters[default_param_name] = default_param_value
            logging.info('parameter {} not specified, using default value: {}'.format(
                                default_param_name, default_param_value)
                            )

    do_reduce(algorithm_name, parameters, version, data_path, reduced_path)



import os
import pkgutil
import inspect
import importlib
import pickle

from vizuka.clustering import (
        kMeans,
        dbscan,
        dummy,
        clusterizer,
        )

from vizuka.plugins import clustering as clustering_plugins

def load_cluster(path):
    with open(path, 'rb') as f:
        return pickle.load(f)

def list_clusterizer():
    built_in_clusterizers = {
            'kmeans':kMeans.KmeansClusterizer,
            'dbscan':dbscan.DBSCANClusterizer,
            'dummy':dummy.DummyClusterizer,
            }

    extra_cluterizers = {}
    
    for (module_loader, name, ispkg) in pkgutil.iter_modules(clustering_plugins.__path__):
        plugin = importlib.import_module('.plugins.clustering.'+name, 'vizuka')
        members = inspect.getmembers(plugin)
        for _, class_ in members:
            if class_ in clusterizer.Clusterizer.__subclasses__():
                extra_cluterizers[name] = class_

    return built_in_clusterizers, extra_cluterizers
    

def make_clusterizer(xs, method='dummy', **kwargs):
    """
    Clusterize the data with specified algorithm
    Naively assume you pass the right parameters for the right algo

    :param data:       array with shape (n,2) of inputs to clusterize
    :param method:     algo to use, supported: kmeans, dbscan, dummy
    :param n_clusters: nb of clusters to find (if applicable)

    :return: a clusterizer object (instance of child of Clusterizer())
    """
    
    built_in_clusterizers, extra_clusterizers = list_clusterizer()
    available_clusterizers = {**built_in_clusterizers, **extra_clusterizers}
    
    clusterizer_builder = available_clusterizers.get(method, None)

    clusterizer = clusterizer_builder(
            required_arguments = {
                param:kwargs[param] for param in clusterizer_builder.required_arguments
                },
            )
    
    clusterizer.fit(xs)
    return clusterizer



import os

import numpy as np

from vizuka.config import (
        VERSION,
        REDUCED_DATA_NAME,
        REDUCED_DATA_PATH,
        )

class Projector():

        def __init__(self):
            """
            Instanciate a Projector engine, you should give it
            here all the parameters
            """
            self.projections = np.array([])
            self.parameters = [] # In your implementation, do add you parameters here !
                                 # If you don't, two projections with different
                                 # parameters will be saved in the same file, the new
                                 # one overwritting the old one ..see: self.save
        
        def project(self, x):
            """
            Projects the x vector into a 2D space according to the
            parameters specified in __init__
            """
            self.projections = np.array([[0,0] for _ in range(len(x))])
            return self.projections

        def get_savename(self, version, filename, path):
            name = ''.join('_' + str(param) for param in self.parameters)
            return ''.join([path, filename, name, '_', version, '.npz'])
     

        def load_projection(self, version=VERSION, base_filename=REDUCED_DATA_NAME, path=REDUCED_DATA_PATH):
            full_path = self.get_savename(version, base_filename, path)
            if os.path.exists(full_path):
                return np.load(full_path)['x_2D']
            else:
                import logging
                logging.info("Nothing found in {}!".format(full_path))
                return np.array([])

        
        def save_projection(self, version=VERSION, base_filename=REDUCED_DATA_NAME, path=REDUCED_DATA_PATH):
            """
            Save the projection in standard location with appropriate name,
            which depends on the parameters of the algorithm
            """
            full_path = self.get_savename(version, base_filename, path)
            np.savez(full_path, x_2D=self.projections)



import os
import pkgutil
import inspect
import importlib
import pickle

from vizuka.similarity import similarity
from vizuka.similarity import bhattacharyya
from vizuka.similarity import deterministic

from vizuka.plugins import similarity as similarity_plugins


def list_similarity():
    built_in_similarity= {
            'bhattacharyya' :bhattacharyya.BhattacharyyaSimilarity,
            'all'           :deterministic.AllFrontiers,
            'none'          :deterministic.NoneFrontiers,
            }

    extra_similarity = {}
    
    for (module_loader, name, ispkg) in pkgutil.iter_modules(similarity_plugins.__path__):
        plugin = importlib.import_module('.plugins.similarity.'+name, 'vizuka')
        members = inspect.getmembers(plugin)
        for _, class_ in members:
            if inspect.isclass(class_):
                if issubclass(class_, similarity.Similarity):
                    extra_similarity[name] = class_

    return built_in_similarity, extra_similarity
    

def make_frontier(method):
    built_in_similarity, extra_similarity = list_similarity()
    available_similarity = {**built_in_similarity, **extra_similarity}
    
    similarity_builder = available_similarity.get(method, None)
    
    return similarity_builder()



"""
Here is the code summoned to reduce the dimension of your
precious data, and also to load it.
We use t-SNE and, if you want, PCA just before it.

..note:: tSNE from sklearn is not the best but is standard
I suggest you to uncomment 'from MulticoreTSNE import TSNE as tsne'
as it will be much faster and won't crash if you use too much RAM.
However this needs extra-install steps :
-> cf https://github.com/DmitryUlyanov/Multicore-TSNE
"""

import itertools
import os
import logging

import numpy as np
from sklearn.decomposition import PCA as PCA_algorithm

from vizuka.config import (
    INPUT_FILE_BASE_NAME,
    DATA_PATH,
    VERSION,
    REDUCED_DATA_PATH,
    REDUCED_DATA_NAME,
)
from vizuka.dimension_reduction import projector


class PCA(projector.Projector):

    def __init__(self, nb_dimension=2, min_ratio_variance_explained=-1):
        """"
        Prepare a PCA projection, to a d-dimensional space or
        with to a space with the minimum number of dimension
        required to have a :param min_ratio_variance_explained:
        ratio of explained variance, if specified.

        :param nb_dimension: the number of dimension to keep, usually
        you want 2
        :param min_ratio_variance_explained: default to -1, if between
        [0,1] it will adjust the number of dimensions for the projection's
        explained variance ratio to match this param
        """
        self.min_ratio_variance_explained = min_ratio_variance_explained
        self.nb_dimension = nb_dimension
        self.parameters = [nb_dimension, min_ratio_variance_explained]
        
        self.engine = PCA_algorithm(svd_solver='randomized')


    def project(self, x):
        """
        Projects :param x:
        ..seealso: __init__
        """
        logging.info(   "starting PCA dimensional reduction,",
                        "needs an explained variance of {}\%".format(
                                self.min_ratio_variance_explained*100)
                        )

        self.engine.fit(x)

        if self.min_ratio_variance_explained:
            variance_explained = 0.
            nb_dim_to_keep     = 0

            while variance_explained < self.min_ratio_variance_explained:
                variance_explained += pca.explained_variance_ratio_[nb_dim_to_keep]
                nb_dim_to_keep += 1
        else:
            nb_dim_to_keep = self.nb_dimension

        x_pcaed          = pca.fit_transform(x)
        self.projections = x_pcaed[:,:nb_dim_to_keep]

        logging.info(   "PCA successfull,",
                        "{} dimensions (axis) where kept after orthnormalization".format(
                                nb_dim_to_keep)
                        )

        return self.projections



import math
import numpy as np

from vizuka.similarity import similarity

class BhattacharyyaSimilarity(similarity.Similarity):

    def __init__(self):
        pass

    def compare(self, dict1, dict2, inside01=True):
        """
        Similarity measure between two empirical distribution
        
        :param dict1: dictionnary containing {class_label:occurence}
        :param dict2: dictionnary containing {class_label:occurence}
        :param inside01: value between 0 and 1, 1 being a complete
        similarity between the two clusters
        """
        s = 0
        max_value_possible = 0
        
        for i in {*dict1, *dict2}:
            s+=(dict1.get(i,0)*dict2.get(i,0))**.5
            if inside01:
                max_value_possible+= (max(dict1.get(i,0),dict2.get(i,0))**.5)
        if inside01:
            return -math.log(s)/max_value_possible if s!=0 else -np.inf
        else:
            return -math.log(s) if s!=0 else -np.inf



"""
This is the main script to launch everything
Is able to reduce the data and launch a vizualization
from it.
"""

import logging

import matplotlib
matplotlib.use('Qt5Agg')  # noqa
import numpy as np
import os
import argparse

from vizuka import dimension_reduction
from vizuka import data_loader
from vizuka import vizualization
from vizuka import launch_reduce


def main():

    """
    See --help if you want help
    """
    
    from vizuka.config import (
        DATA_PATH,
        VERSION,
        REDUCED_DATA_PATH,
        REDUCED_DATA_NAME,
        MODEL_PATH,
        INPUT_FILE_BASE_NAME,
        RAW_NAME,
        PROJECTION_DEFAULT_PARAMS,
        DEFAULT_PROJECTOR,
    )


    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--mnist', action='store_true',
         help='download, fit, and vizualize MNIST dataset')
    parser.add_argument(
            '--show-required-files', action='store_true',
            help='show all the required files for optimal use',
            )
    parser.add_argument(
        '-p', '--path',
         help='change the location of your data/ folder, containing set/ reducted/ graph/ models/')
    parser.add_argument(
        '-f', '--feature_to_filter', action='append',
         help='Adds a feature listed in raw_data.npz["columns"] to the filters list')
    parser.add_argument(
        '-s', '--feature_to_show', action='append',
        help='Usage : -s MY_COLUMN_NAME:PLOTTER with PLOTTER being a value in %. Adds this non-preprocessed/human-readable feature to the cluster view'.format())
    parser.add_argument(
        '-r', '--reduce', action="store_true",
         help='launch a full dimension reduction')
    parser.add_argument(
        '-h1', '--heatmap1',
         help='Specify the 1st heatmap to show')
    parser.add_argument(
        '-h2', '--heatmap2',
         help='Specify the 2nd heatmap to show')
    parser.add_argument(
        '--use_pca', 
         help='force a PCA dimensional reduction, needs a minimum variance ratio explained')
    parser.add_argument(
        '-v', '--version', type=str,
        help='(optional) specify a version of the files to load/generate (see vizuka --show_required_files), currently: '+VERSION)
    parser.add_argument(
        '--force-no-predict', action="store_true",
        help='(not recommended) do not load a predictions file : vizualize as if you predicted with 100\% accuracy')
    parser.add_argument(
        '--no-vizualize', action="store_true",
         help='(for debug) do not prepare a nice data vizualization')
    parser.add_argument(
        '--no-plot', action="store_true",
         help='(for debug) do not show a nice data vizualization (but prepare it nonetheless)')
    
    parser.set_defaults(
            heatmap1='accuracy',
            heatmap2='entropy',
            no_plot=False, 
            no_vizualize=False,
            show_required_files=False,
            version=VERSION,
            path=os.path.dirname(__file__),
            feature_to_filter=[],
            feature_to_show=[],
            use_pca = 0,
            force_no_predict = False,
            mnist=False,
            )

    args = parser.parse_args()

    MODEL_PATH = os.path.join(args.path, MODEL_PATH)
    INPUT_FILE_BASE_NAME = os.path.join(args.path, INPUT_FILE_BASE_NAME)
    REDUCED_DATA_PATH = os.path.join(args.path, REDUCED_DATA_PATH)
    DATA_PATH = os.path.join(args.path, DATA_PATH)

    reduce_      = args.reduce
    no_vizualize = args.no_vizualize
    no_plot      = args.no_plot
    version      = args.version
    features_name_to_filter  = args.feature_to_filter
    features_name_to_display = args.feature_to_show
    pca_variance_needed = args.use_pca
    force_no_predict = args.force_no_predict
    show_required_files=args.show_required_files
    heatmap1    = args.heatmap1
    heatmap2    = args.heatmap2

    if args.show_required_files:
        print(
        'VERSION: string that identifies your dataset (default is MNIST_example)\n\n'
        '\nVizuka needs the following files :\n\n'
        '\t + data/set/preprocessed_inputs_VERSION.npz\n'
        '\t ------------------------------------------\n'
        '\t\t x:\t preprocessed inputs\n'
        '\t\t y:\t outputs to be predicted\n'
        '\t\t NB:\t this is the only mandatory file, the following is highly recommended:\n'
        '\n\n'
        '\t + data/models/predict_VERSION.npz -> optional but recommended\n'
        '\t -------------------------------------------------------------\n'
        '\t\t pred:\t predictions returned by your algorithm\n'
        '\t\t NB:\t should be same formatting as in preprocessed_inputs_VERSION["y"])\n'
        '\n\n'
        '\t + raw_data.npz -> optional\n'
        '\t --------------------------\n'
        '\t\t x:\t\t array of inputs BEFORE preprocessing\n'
        '\t\t\t\t\t probably human-readbable, thus useful for vizualization\n'
        '\t\t columns:\t the name of the columns variable in x\n'
        '\t\t NB:\t this file is used if you run vizuka with\n'
        '\t\t\t    --feature-name-to-display COLUMN_NAME:PLOTTER COLUMN_NAME2:PLOTTER2 or\n'
        '\t\t\t    --feature-name-to-filter COLUMN_NAME1 COLUMN_NAME2 (see help for details)\n'
        '\n\n'
        '\t + reduced/2Dembedding_PARAMS_VERSION.npz -> reaaaally optional\n'
        '\t --------------------------------------------------------------\n'
        '\t\t x2D:\t projections of the preprocessed inputs x in a 2D space\n'
        '\t\t NB:\t this set is automatically generated with tSNE but you can specify your own\n'
        )
        return

    
    if args.mnist:
        from vizuka.example import load_mnist
        version = load_mnist.version
        features_name_to_display = ['image:images']

    
    new_fntd = {}
    error_msg  = 'Argument should be feature_name:plotter with plotter'
    error_msg += '\nin (logdensity, density, wordcloud, counter)'
    e = Exception(error_msg)
    for feature_name in features_name_to_display:
        if ':' not in feature_name:
            raise e
        k,v = feature_name.split(':')
        # TODO if v not in cluster_diving.plotter.keys():
        #    raise e
        plotters = new_fntd.get(k, [])
        plotters.append(v)
        new_fntd[k] = plotters
    features_name_to_display = new_fntd

    logging.info("Starting script")
    logging.info("raw_data=loading")
    (
        x,
        y,
        class_encoder,
        class_decoder,
    ) = data_loader.load_preprocessed(
            file_base_name   = INPUT_FILE_BASE_NAME,
            path             = DATA_PATH,
            version          = version,
            )

    logging.info('raw_data=loaded')

    x_2D = []
    if not reduce_:
        x_2D = data_loader.load_projection(
            algorithm_name        = DEFAULT_PROJECTOR,
            parameters            = PROJECTION_DEFAULT_PARAMS[DEFAULT_PROJECTOR],
            version               = version,
            path                  = REDUCED_DATA_PATH,
            base_filename         = REDUCED_DATA_NAME,
        )

    if not x_2D.size:
        logging.info("no reduced data found! Needs to learn some dimension reduction..")
        force_reduce = True
    else:
        force_reduce = False

    if force_reduce or reduce_: # if nothing loaded or reduce is forced by arg

        x_2D = launch_reduce.do_reduce(
            algorithm_name          = DEFAULT_PROJECTOR,
            parameters              = PROJECTION_DEFAULT_PARAMS[DEFAULT_PROJECTOR],
            version                 = version,
            data_path               = DATA_PATH,
            reduced_path            = REDUCED_DATA_PATH,
        )
    
    ###############
    # PREDICT
    if force_no_predict:
        x_predicted = y
    else:
        try:
            logging.info('predictions=loading')
            x_predicted = data_loader.load_predict(
                path=MODEL_PATH,
                version=version,
            )
            logging.info('RNpredictions=ready')
        except FileNotFoundError:
            logging.info((
                    "Nothing found in {}, no predictions to vizualize\n"
                    "if this is intended you can force the vizualization"
                    "with --force_no_predict :\n{}\n"
                    ).format(MODEL_PATH, os.listdir(MODEL_PATH)))
            return
    
    raw = data_loader.load_raw(version, DATA_PATH)
    if raw:
        logging.info("loading raw transactions for analysis..")
        raw_data, raw_columns = raw
    else:
        logging.info('no raw data provided, all cluster vizualization disabled! (-s and -f options)')
        raw_data    = []
        raw_columns = []
        features_name_to_filter  = []
        features_name_to_display = {}


    
    if not no_vizualize:

        f = vizualization.Vizualization(
            projected_input=x_2D,
            predicted_outputs=x_predicted,
            raw_inputs=raw_data,
            raw_inputs_columns=raw_columns,
            correct_outputs=y,
            resolution=200,
            class_decoder=class_decoder,
            class_encoder=class_encoder,
            special_class='x',
            nb_of_clusters=120,
            features_name_to_filter  = features_name_to_filter,
            features_name_to_display = features_name_to_display,
            heatmaps_requested = [heatmap1, heatmap2],
            output_path=os.path.join('output.csv'),
            model_path=MODEL_PATH,
            version=version,
        )

        if not no_plot:
            f.plot()
            f.show()
    return f

if __name__ == '__main__':
    main()



"""
Qt handler for drawing buttons and IHM black magic.
Please do read it you fool. I must admit I am not
proud of everything written down there.

This should be rewritten with QtCreator's help.
"""
import matplotlib
import sys
import logging
import os

matplotlib.use('Qt5Agg')  # noqa
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
import matplotlib.gridspec as gridspec

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import (
    QApplication,
    QMainWindow,
    QWidget,
    QHBoxLayout,
    QVBoxLayout,
    QSizePolicy,
    QDockWidget,
    QLineEdit,
    QInputDialog,
)

from vizuka.graphics import qt_helpers
from vizuka.cluster_diving import moar_filters
from vizuka import clustering
from vizuka import similarity
from vizuka.cluster_viewer import make_plotter


def onclick_wrapper(onclick):
    """
    This decorator for onclick detects if the mouse event
    should trigger something.
    I tried to make it worst but this is my best
    """
    def wrapper(*args, **kwargs):
        if args[0].detect_mouse_event:  # args[0] is self
            return onclick(*args, **kwargs)
        else:
            return lambda x: None
    return wrapper

CLUSTER_PLOTTER = {}
            
class Cluster_viewer(matplotlib.figure.Figure):

    def __init__(self, features_to_display, x_raw, x_raw_columns, show_dichotomy=True):
        super().__init__()
        self.x_raw = x_raw
        self.x_raw_columns = x_raw_columns
        self.show_dichotomy = show_dichotomy # deprecated

        self.features_to_display = features_to_display
        self.spec_by_name = {}
        self.cluster_view_selected_indexes = []

        self.spec = gridspec.GridSpec(
                len(features_to_display.keys()),
                2, wspace=0.2
                )
        
        for idx,feature_name in enumerate(features_to_display.keys()):
            for plotter in features_to_display[feature_name]:
                self.spec_by_name[feature_name+plotter] = {}
                self.spec_by_name[feature_name+plotter]['good'] = self.spec[idx%2]
                self.spec_by_name[feature_name+plotter]['bad' ] = self.spec[idx%2+1]

                if plotter not in CLUSTER_PLOTTER.keys():
                    CLUSTER_PLOTTER[plotter] = make_plotter(plotter)
       

    def update_cluster_view(self, clicked_cluster, index_by_cluster_label, indexes_good, indexes_bad):
        """
        Updates the axes with the data of the clicked cluster

        clicked cluster: the label of the cluster you clicked
        index_by_cluster_label: indexs of datas indexed by cluster label (set containing int)
        indexes_good: indexes of all good predictions
        indexes_bad: indexes of all bad predicitons
        """
        self.clear()

        self.cluster_view_selected_indexes += index_by_cluster_label[clicked_cluster]

        selected_xs_raw  ={'all': [self.x_raw[idx] for idx in self.cluster_view_selected_indexes]}
        if self.show_dichotomy:
            selected_xs_raw['good'] = [self.x_raw[idx] for idx in self.cluster_view_selected_indexes if idx in indexes_good]
            selected_xs_raw['bad' ] = [self.x_raw[idx] for idx in self.cluster_view_selected_indexes if idx in indexes_bad ]
        
        columns_to_display = [list(self.x_raw_columns).index(i) for i in self.features_to_display]
        data_to_display = {
                'all':
                        {
                        self.x_raw_columns[i]:[x[i] for x in selected_xs_raw['all']]
                        for i in columns_to_display
                        }
                    }
        if self.show_dichotomy:
            data_to_display['good'] = {
                self.x_raw_columns[i]:[x[i] for x in selected_xs_raw['good']]
                for i in columns_to_display
                }
            data_to_display['bad'] = {
                self.x_raw_columns[i]:[x[i] for x in selected_xs_raw['bad']]
                for i in columns_to_display
                }

        def plot_it(plotter, data_to_display, data_name, fig, spec_to_update_, key):

            spec_to_update = spec_to_update_[key]
            data = data_to_display[key][data_name]
            axe = plotter(data, fig, spec_to_update)
            if 'log' in data_to_display[key][data_name]:
                data_name += ' - log'
            data_name +=  ' - {} predictions'.format(key)

            if axe:
                axe.set_title(data_name)


        for key in ['good', 'bad']:
            for data_name in self.features_to_display:
                for plotter_name in self.features_to_display[data_name]:
                    plotter = CLUSTER_PLOTTER[plotter_name]
                    spec_to_update = self.spec_by_name[data_name+plotter_name]
                    plot_it(plotter, data_to_display, data_name, self, spec_to_update, key) 
        
    def reset(self):
        self.clear()
        self.cluster_view_selected_indexes = []


class Qt_matplotlib_handler():

    def __init__(self, figure):
        self.right_dock = QtCore.Qt.RightDockWidgetArea
        self.textboxs = {}
        self.plottings = []

        # configure the app + main window
        self.app = QApplication(sys.argv)
        self.window = QMainWindow()
        self.features_to_filter=None
        self.figure = figure


    def show(self):
        """
        Shows the window you built with your tears
        """
        self.window.showMaximized()
        for win in self.additional_windows:
            win.show()
        self.refresh()
        logging.info("refreshed")
        logging.info("showed")

    def refresh(self):
        """
        Refresh the matplotlib plots
        """
        for p in self.plottings:
            p.canvas.draw()
                
    
    @onclick_wrapper
    def onclick(self, *args, **kwargs):
        """
        ..seealso::onclick_wrapper
        """
        self.base_onclick(*args, **kwargs)

        
    def toogle_detect_mouse_event(self, *args, **kwargs):
        """
        Enable/Disable mouse_event handling ..seealso::onclick_wrapper
        """
        self.detect_mouse_event = not self.detect_mouse_event


class Viz_handler(Qt_matplotlib_handler):
    """
    A Viz_handler is attached to the viz_engine defined in vizualization.py
    It basically lists lots of QWidget things and propose methods to init
    them "gracefully", hum...

    Only IHM here.
    """
    
    def __init__(
            self,
            viz_engine,
            figure,
            onclick,
            additional_filters=[],
            additional_figures=[],
            ):
        """
        This object is a QtWindow (or 2-3-4...) with a mamtplotlib.Figure
        and a onclick event handler. As you can see it is also linked to
        a viz_engine which will do the magic.
        
        This instance should only do IHM stuff, nothing intellectual, everything
        is handled by the viz_engine
        """
        super(Viz_handler, self).__init__(figure)

        self.viz_engine = viz_engine
        self.detect_mouse_event = False
        self.base_onclick = onclick
        self.window.setWindowTitle('Data vizualization')
        self.additional_filters = additional_filters
        self.additional_figures = additional_figures

        self.additional_windows = []

        # add the main figure
        qt_helpers.add_figure(self.figure, window=self.window, plottings=self.plottings, onclick=self.onclick)
        #self.cluster_diver = Cluster_diver(self.x_raw, self.x_raw_columns, ['ape_code'])

        # add additional window
        if self.additional_filters:
            self.additional_windows.append(qt_helpers.add_window(self.window, 'moar filters'))
            moar_filters(
                    window=self.additional_windows[-1],
                    right_dock=QtCore.Qt.RightDockWidgetArea,
                    features = self.viz_engine.x_raw,
                    all_features_categories = self.viz_engine.x_raw_columns,
                    features_to_filter=additional_filters,
                    viz_engine=self.viz_engine,
                    )

        for fig in self.additional_figures:
            self.additional_windows.append(qt_helpers.add_window(self.window, 'Cluster viewer'))
            qt_helpers.add_figure(fig, self.additional_windows[-1], plottings=self.plottings)


        """
        if self.features_to_display:        
            self.cluster_window = add_window(self.window, 'Cluster view')
            add_cluster_view(
                    window=self.cluster_window,
                    right_dock = QtCore.Qt.RightDockWidgetArea,
                    features = self.viz_engine.x_raw,
                    all_features_categories = self.viz_engine.x_raw_columns,
                    features_to_diplay=features_to_display,
                    viz_engine=self.viz_engine,
        """


        # self.add_figure(self.additional_figure, onclick=None, window=self.additional_window)

        # logging.info("textboxs=adding")
        qt_helpers.add_checkboxes(
            self.window,
            "Filter by true class",
            self.viz_engine.possible_outputs_list,
            self.viz_engine.filter_by_correct_class,
            self.right_dock,
        )
        qt_helpers.add_checkboxes(
            self.window,
            "Filter by predicted class",
            self.viz_engine.possible_outputs_list,
            self.viz_engine.filter_by_predicted_class,
            self.right_dock,
        )
        qt_helpers.add_checkboxes(
            self.window,
            "Navigation options",
            ['detect mouse event'],
            self.toogle_detect_mouse_event,
            self.right_dock,
        )
        # logging.info("textboxs=ready")

        # add button
        # logging.info("action buttons=adding")
        #if self.viz_engine.x_raw:
        qt_helpers.add_button(
            self.window,
            "Export x",
            self.export,
            self.right_dock,
        )
        qt_helpers.add_button(
            self.window,
            "Save clusterization",
            self.save_clusters,
            self.right_dock,
        )

        # add menulist
        builtin_cl, extra_cl = clustering.list_clusterizer()
        self.available_clustering_engines = {**builtin_cl, **extra_cl}
        qt_helpers.add_menulist(
            self.window,
            'Clustering method',
            'Clusterize', [*self.available_clustering_engines.keys()],
            self.request_new_clustering,
            dockarea=self.right_dock,
        )

        builtin_fr, extra_fr = similarity.list_similarity()
        qt_helpers.add_menulist(
            self.window,
            'Clusters borders',
            'Delimits',
            [*builtin_fr.keys(), *extra_fr.keys()],
            self.viz_engine.request_new_frontiers,
            self.right_dock,
        )
        qt_helpers.add_menulist(
            self.window,
            'Predictor set',
            'Load',
            self.viz_engine.predictors,
            self.viz_engine.reload_predict,
            self.right_dock,
        )
        self.user_cluster_menulist = qt_helpers.add_menulist(
            self.window,
            'Saved clusters',
            'Load',
            self.viz_engine.saved_clusters,
            self.viz_engine.load_clusterization,
            self.right_dock,
        )
    def save_clusters(self):
        text, validated = QInputDialog.getText(
                self.window,
                "YOLO", "Name for the clusters you selected ?"
                )
        if validated:
            if text == '':
                text = "clusters"
            self.viz_engine.save_clusterization(text+'.pkl')

    def export(self):
        text, validated = QInputDialog.getText(
                self.window,
                "YOLO", "Name for the exported csv ?"
                )
        if validated:
            if text == '':
                text = "export.csv"
        self.viz_engine.export(text)

    def request_new_clustering(self, clustering_engine_name):
        self.clustering_params={}
        for requested_param in self.available_clustering_engines[clustering_engine_name].required_arguments:
            text, validated = QInputDialog.getText(
                    self.window,
                    "YOLO", requested_param+' ?'
                    )
            self.clustering_params[requested_param]=float(text)
        return self.viz_engine.request_new_clustering(clustering_engine_name)



class Heatmap():

    def __init__(self, **kwargs):
        """
        Here you create and instanciate the heatmap
        Any argument may be passed
        """
        self.update_colors(kwargs)

    def update_colors(self, **kwargs):
        """
        Update your heatmap with any argument provided
        self.all_colors must be a dict indexed by x by y
        """
        self.all_colors = {}

    def get_all_colors(self):
        return self.all_colors



import os
import pkgutil
import inspect
import importlib
import pickle

from vizuka.heatmap import heatmap
from vizuka.heatmap import accuracy
from vizuka.heatmap import entropy

from vizuka.plugins import heatmap as heatmap_plugins


def list_heatmap():
    """
    List all the available heatmaps
    First it looks in this directory, then in plugin/heatmaps/
    All heatmaps must inherit vizuka.heatmap.heatmap.Heatmap to be detectable
    
    :return: a dict with built-in heatmaps and their respective constructor
             and a dict with plugins heatmaps and their respective constructor
    """
    built_in_heatmap= {
            'accuracy' : accuracy.AccuracyHeatmap,
            'entropy'  : entropy.EntropyHeatmap,
            }

    extra_heatmap = {}
    
    for (module_loader, name, ispkg) in pkgutil.iter_modules(heatmap_plugins.__path__):
        plugin = importlib.import_module('.plugins.heatmap.'+name, 'vizuka')
        members = inspect.getmembers(plugin)
        for _, class_ in members:
            if inspect.isclass(class_):
                if issubclass(class_, heatmap.Heatmap):
                    extra_heatmap[name] = class_

    return built_in_heatmap, extra_heatmap
    

def make_heatmap(method):
    """
    Gets a list of all available heatmap constructors and returns the correct one

    :param method: the name of the heatmap you want
    """
    built_in_heatmap, extra_heatmap = list_heatmap()
    available_heatmap = {**built_in_heatmap, **extra_heatmap}
    
    heatmap_builder = available_heatmap.get(method, None)
    
    return heatmap_builder



class Similarity():

    def __init__(self):
        pass

    def compare(self, dict1, dict2):
        """
        :param dict1: a dictionnary containing {class_name:occurence}
        :param dict2: same

        :return: a similarity score between 0 and 1, 1 being the maximum similarity value
        (ideally 1 if and only if dict1==dict2)
        """
        return 0



import logging
import os
import itertools

import numpy as np

from vizuka import dimension_reduction
from vizuka.config import (
        MODEL_PATH,
        VERSION,
        DEFAULT_PREDICTOR,
        PROJECTION_DEFAULT_PARAMS,
        REDUCED_DATA_PATH,
        REDUCED_DATA_NAME,
        INPUT_FILE_BASE_NAME,
        DATA_PATH,
        RAW_NAME,
        )


def load_predict(path=MODEL_PATH, version=VERSION, namePredictor=DEFAULT_PREDICTOR):
    """
    Simply load the predictions associated with the VERSION data
    """
    logging.info("trying to load {}".format(path + namePredictor +'_'+ version + '.npz'))
    return np.load(path + namePredictor + version + '.npz')['pred']

def load_predict_byname(filename, path=MODEL_PATH):
    """
    Simply load the predictions associated with the VERSION data
    """
    full_path = os.path.join(path, filename)
    logging.info("trying to load {}".format(full_path))
    return np.load(os.path.join(path, filename))['pred']

def load_projection(algorithm_name, parameters, version, base_filename, path):
    logging.info("data_loader=loading projection:\n\talgorithm:{}\n\tparameters:{}".format(
                        algorithm_name,
                        parameters,
                        ))
    algo_builder = dimension_reduction.make_projector(algorithm_name)
    algo = algo_builder(**parameters)
    projection = algo.load_projection(version=version, base_filename=base_filename, path=path)

    logging.info("data_loader=ready")
    return projection
    
def load_tSNE(
        params=PROJECTION_DEFAULT_PARAMS['tsne'],
        version=VERSION,
        path=REDUCED_DATA_PATH,
        ):
    """
    Load tSNE representation.

    :param params: dict of multiples t-SNE parameters of the representations
    we want
                   load every combination available
                   .. seealso:: sklearn.manifold.TSNE()
    :type params: {
                    'perplexities':array(int),
                    'learning_rates':array(int),
                    'n_iter':array(int),
                    }
    :param version: version of data to load (e.g: _20170614)
    :param path: location of the 2D representation to load
    is divided (tsne is greedy)

    :return: Embedded data in 2D space, and t-SNE model
    :rtype:  dict{params:(float,float)}, dict({params:tsne.model})
    """

    perplexities = params['perplexities']
    learning_rates = params['learning_rates']
    n_iters = params['n_iters']

    x_transformed = {}
    models = {}

    for perplexity, learning_rate, n_iter in itertools.product(
            perplexities, learning_rates, n_iters):

        param = (perplexity, learning_rate, n_iter)
        name = ''.join('_' + str(p) for p in param)
        full_path = ''.join([
            path,
            REDUCED_DATA_NAME,
            name,
            '_',
            version,
            '.npz',
        ])
        logging.info("embedded data= loading %s %s %s %s from %s",
                     str(perplexity),
                     str(learning_rate),
                     str(n_iter),
                     full_path,
                     )
        if os.path.exists(full_path):
            x_transformed[param] = np.load(full_path)['x_2D']
        
            try:
                models[param] = np.load(full_path)['model']
            except KeyError:
                logging.info("old version, model not found, only embedded data")
            logging.info("embedded data=ready")
        else:
            logging.info("emebedded data = model {} {} {}  not found".format(
                perplexity,
                learning_rate,
                n_iter
                )
            )
    return x_transformed, models

def load_raw(version, path):
    raw_filename = os.path.join(path, RAW_NAME + version + '.npz')
    if os.path.exists(raw_filename):
        raw_ = np.load(raw_filename)
        return raw_["x"], raw_["columns"]
    else:
        return None

def load_preprocessed(
        file_base_name=INPUT_FILE_BASE_NAME,
        path=DATA_PATH,
        version=VERSION,
):
    """
    Loads and returns the data for tSNE
    One-hotted and regular ML-encoding

    File to load should :
        - be in path
        - has name "(INPUT_FILE_BASE_NAME)_x_y_(VERSION).npz"
                (VERSION) is for e.g "_20170614"
                (FILE_BASE_NAME) if for e.g 'processed_predictions' or 'one_hot'
        - contains entry x with input data
        - contains entry y with output data (possible_outputs_list to predict)
        - optionnaly an encoder to translate machine-readable possible_outputs_list to human-readable possible_outputs_list
                (actually it is the opposite e.g: {604600:[False, True, False, .., False]})

    :return: (input for t-SNE, classes, encoder (humantomachine class possible_outputs_list),
    decoder (machinetohuman possible_outputs_list))

    Note that encoder/decoder are assumed trivial if no encoder are found in npz

    """
    xy = np.load(path + INPUT_FILE_BASE_NAME + version + '.npz')
    x, y = xy['x'], xy['y']
    
    class_encoder = {y: y for y in set(y)}
    class_decoder = lambda x: x # noqa

    x = xy['x']
    if 'y' in xy.keys():
        y = xy['y']
        del xy
        return (np.array(x), np.array(y), class_encoder, class_decoder)
    elif 'y' + '_decoded' in xy.keys():
        y_decoded = xy['y' + '_decoded']
        del xy
        return (np.array(x), np.array(y_decoded), class_encoder, class_decoder)




'''
Clustering engine to use with Vizualization

If you want to implement one, dont forget to add it
on qt_handler, to be able to select on the IHM
Also, adds it in make_clusterizer obviously

3 methods are necessary to implement, cf Clusterizer():
    init    - set main params
    fit     - to prepare the algo for the data
    predict - to find out the cluster of a (x,y)
'''
import pickle

from matplotlib import pyplot as plt
import logging


class Clusterizer():

        required_arguments = []
        # List of parameters needed for the clustering engine (asked in dialog popup)

        def __init__(self, required_arguments={}):
            """
            Builds a clusterizer object e.g: kmeans
            Do not know the datas at this point, just pass it the bare
            minimum to initialize an engine.

            :param *args **kwargs: parameters passed to the clusterizer engine
                                   it can literally be whatever you want
            """
            self.engine = None
            self.method=''
            self.required_arguments = required_arguments

        def fit(self, xs):
            """
            First time the engine sees the data.
            Depending on the algorithm you may do your magic your own way
            and maybe store new variable in self, maybe store all the
            predicts for each x directly in a dict.

            :param xs: a list containing data to clusterize
            """
            pass


        def predict(self, xs):
            """
            Finds the cluster(s) in which the data is.

            :params xs: array-like of (x,y) points
            :return: array-like of cluster id
            """
            return (0, ) * len(xs)
        

        def save_cluster(self, path):
            with open(path, 'wb') as f:
                pickle.dump(self, f)




if __name__ == '__main__':
    """
    Yes, I like to test my code with __main__
    """
    from vizuka import dim_reduction as dr
    from vizuka import data_loader as dl
    datas_sets, models = dl.load_tSNE()
    datas = datas_sets[50, 1000, 'pca', 15000]

    clusterizer = make_clusterizer(datas, method='kmeans', n_clusters=80)
    f, ax = plot_clusters(datas, clusterizer)

    plt.show()



from collections import Counter

import numpy as np
import matplotlib
matplotlib.use('Qt5Agg')  # noqa
from matplotlib import pyplot as plt

from vizuka.cluster_viewer.plotter import Plotter

class Counter(Plotter):

    def __call__(self, data, fig, spec):
        """
        Show a counter of each class occurrences

        :param: data is a list of the observations
        :param: fig is the Figure on which the axe will be drawn
        :param:spec is the GridSpec for the axe

        :return: the axe with the visualization
        """
        if not data:
            return
        axe = plt.Subplot(fig, spec)

        c = Counter(data)
        x = [l for l in c.keys()]
        y = [c[l] for l in x]

        
        order = np.argsort(y)
        y = [y[i] for i in order]
        x = [x[i] for i in order]

        graduation = np.linspace(0, len(y), len(y))
        axe.bar(height=y, left=graduation)
        axe.set_xticks(graduation)
        axe.set_xticklabels([str(i) for i in x])
        fig.add_subplot(axe)

        return axe



# HOW TO ADD A CUSTOM HEATMAP ?
###############################

# Defines here a heatmap, its name will be the .py name, and request it at vizuka launch
# (vizuka --heatmap1=my_custom_heatmap_name)
# It should inherits vizuka.heatmap.heatmap.Heatmap (yea sorry 'bout that) and implements the
# basics methods specified there( very simple). See example below or Abstract class in
# vizuka.heatmap.heatmap
#
# One color by cluster !
#
# The object vizuka.Vizualization is passed in the constructor by default, you can change this
# in vizuka.vizualization.request_heatmap() but all the info you may want should be fetchable
# in this great big object
#
# Here is a working example, which is simply the accuracy heatmap :
#
###############################################################################################

#
# import numpy as np
# import logging
# 
# from vizuka.heatmap import heatmap
# 
# class AccuracyHeatmap(heatmap.Heatmap):
# 
#     def __init__(self, vizualization):
#         """
#         Prepare an 'accuracy' heatmap (good predictions / total effectif)
# 
#         Heatmap showing the accuracy of the prediction, 3 colors are actually used :
#             - red for bad prediction
#             - blue for correct
#             - green for special_class which is a special label defined at the Vizualization.__init__
#             (typically the label "0")
#
#         All colors are mixed linearly
#
#         :returns: a dict of colors to be plotted (indexed by x by y)
#
#         """
#         self.update_colors(vizualization)
# 
#     
#     def update_colors(self, vizualization):
#         """
#         Updates the colors with the new vizualization object
#         '''
#         self.all_colors = [[0 for _ in range(vizualization.resolution)] for _ in range(vizualization.resolution) ]
#         centroids_cluster_by_index = vizualization.clusterizer.predict(vizualization.mesh_centroids)
#         logging.info('heatmap: drawing accuracy heatmap')
# 
#         for index, (x, y) in enumerate(vizualization.mesh_centroids):
# 
#             current_centroid_cluster_label = centroids_cluster_by_index[index]
# 
#             nb_good_points = vizualization.nb_good_point_by_cluster.get(current_centroid_cluster_label, 0)
#             nb_bad_points  = vizualization.nb_bad_point_by_cluster.get (current_centroid_cluster_label, 0)
#             nb_null_points = vizualization.nb_null_point_by_cluster.get(current_centroid_cluster_label, 0)
# 
#             nb_of_valid_cluster_points = nb_good_points + nb_bad_points
# 
#             if nb_of_valid_cluster_points > 0:
#                 accuracy_correct   = nb_good_points / float(nb_of_valid_cluster_points)
#                 accuracy_null      = nb_null_points / float(nb_of_valid_cluster_points)
#                 accuracy_incorrect = 1 - accuracy_correct
#             else:
#                 accuracy_correct = 1
#                 accuracy_null = 1
#                 accuracy_incorrect = 1
# 
#             red   = accuracy_incorrect
#             green = accuracy_null
#             blue  = accuracy_correct
#             
#             x_coordinate, y_coordinate = vizualization.get_coordinates_from_index(index)
# 
#             self.all_colors[x_coordinate][y_coordinate ] = [red, green, blue]
# 
#     def get_all_colors(self):
#         """
#         Returns colors by x by y coordinates
#         """
#         return self.all_colors
# 
#
########################################################                                                        



from vizuka.similarity import similarity

class AllFrontiers(similarity.Similarity):
    def __init__(self):
        pass
    def compare(dict0, dict1, force_inside01=True):
        """
        Draw all frontiers
        """
        return 0

class NoneFrontiers(similarity.Similarity):
    def __init__(self):
        pass
    def compare(dict0, dict1, force_inside01=True):
        """
        Draw none frontiers
        """
        return 1




import matplotlib
matplotlib.use('Qt5Agg')  # noqa
from matplotlib import  pyplot as plt

class Plotter(object):
    """
    Callable function that returns an axe with a visualization
    Cf vizuka/.cluster_viewer.counter for a simple example
    """

    def __call__(self, data, fig, spec):
        """
        :param: data is a list of the observations
        :param: fig is the Figure on which the axe will be drawn
        :param:spec is the GridSpec for the axe

        :return: the axe with the visualization
        """
        axe = plt.Subplot(fig, spec)
        fig.add_subplot(axe)
        return axe



"""
This module provides heatmaps to be drawn on the side of the main map.
Basically it takes a vizualization object and returns an array with 
associated colors.

If you want to provide new ones dont forget to register them in qt_handler
"""

import numpy as np
import logging

from vizuka.heatmap import heatmap

class AccuracyHeatmap(heatmap.Heatmap):

    def __init__(self, vizualization):
        """
        Prepare an 'accuracy' heatmap (good predictions / total effectif). One color by cluster !

        Heatmap showing the accuracy of the prediction, 3 colors are actually used :
            - red for bad prediction
            - blue for correct
            - green for special_class which is a special label defined at the Vizualization.__init__
            (typically the label "0")

        All colors are mixed linearly

        :returns: a dict of colors to be plotted (indexed by x by y)

        """
        self.update_colors(vizualization)

    
    def update_colors(self, vizualization):
        """
        Updates all the colors from a new vizualization object (containing all infos about
        predictions, true class, clusters, etc...)
        """
        self.all_colors = [[0 for _ in range(vizualization.resolution)] for _ in range(vizualization.resolution) ]
        centroids_cluster_by_index = vizualization.clusterizer.predict(vizualization.mesh_centroids)
        logging.info('heatmap: drawing accuracy heatmap')

        for index, (x, y) in enumerate(vizualization.mesh_centroids):

            current_centroid_cluster_label = centroids_cluster_by_index[index]

            nb_good_points = vizualization.nb_good_point_by_cluster.get(current_centroid_cluster_label, 0)
            nb_bad_points  = vizualization.nb_bad_point_by_cluster.get (current_centroid_cluster_label, 0)
            nb_null_points = vizualization.nb_null_point_by_cluster.get(current_centroid_cluster_label, 0)

            nb_of_valid_cluster_points = nb_good_points + nb_bad_points

            if nb_of_valid_cluster_points > 0:
                accuracy_correct   = nb_good_points / float(nb_of_valid_cluster_points)
                accuracy_null      = nb_null_points / float(nb_of_valid_cluster_points)
                accuracy_incorrect = 1 - accuracy_correct
            else:
                accuracy_correct = 1
                accuracy_null = 1
                accuracy_incorrect = 1

            red   = accuracy_incorrect
            green = accuracy_null
            blue  = accuracy_correct
            
            x_coordinate, y_coordinate = vizualization.get_coordinates_from_index(index)

            self.all_colors[x_coordinate][y_coordinate ] = [red, green, blue]

        logging.info('heatmap: accuracy done')
        
    def get_all_colors(self):
        """
        Returns colors by x by y coordinates
        """
        return self.all_colors





import os
import pkgutil
import inspect
import importlib
import pickle

from vizuka.dimension_reduction import (
        projector,
        pca,
        tsne,
        )

from vizuka.plugins import dimension_reduction as projector_plugins

def list_projectors():
    built_in_projectors = {
            'pca':pca.PCA,
            'tsne':tsne.tSNE,
            }

    extra_projectors= {}
    
    for (module_loader, name, ispkg) in pkgutil.iter_modules(projector_plugins.__path__):
        plugin = importlib.import_module('.plugins.dimension_reduction.'+name, 'vizuka')
        members = inspect.getmembers(plugin)
        for _, class_ in members:
            if class_ in projector.Projector.__subclasses__():
                extra_projectors[name] = class_

    return built_in_projectors, extra_projectors
    

def make_projector(method='tsne', **kwargs):
    """
    Return a projector function

    :param method: the name of the algo
    """
    
    built_in_projectors, extra_projectors = list_projectors()
    available_projectors= {**built_in_projectors, **extra_projectors}
    
    projector_builder = available_projectors.get(method, None)
    
    return projector_builder



from setuptools import setup, find_packages
import os

data = {
        'vizuka':
        [
            'data/set/.gitkeep',
            'data/models/.gitkeep',
            'data/cache/.gitkeep',
            'data/saved_clusters/.gitkeep',
            'data/reduced/.gitkeep',
            'example/2Dembedding1_50_1000_random_12000_MNIST_example.npz',
        ],
    }

setup(
        name='Vizuka',
        version='0.29.1',
        packages=find_packages(),#['vizuka/'],
        package_data = data,
        entry_points = {
            'console_scripts':[
                'vizuka=vizuka.launch_viz:main',
                'vizuka-reduce=vizuka.launch_reduce:main',
                ],
            },
        description='Represents your high-dimensional datas in a 2D space and play with it',
        long_description = open(os.path.join(os.path.dirname(__file__),'README.md')).read(),
        install_requires = open(os.path.join(os.path.dirname(__file__), 'vizuka/requirements/requirements.txt')).read(),
        license = 'GPL V3',
        author='Sofian Medbouhi',
        author_email='sof.m.sk@free.fr',
        )
            



import numpy as np
import logging

from vizuka.helpers.ml_helper import cross_entropy
from vizuka.heatmap import heatmap


class EntropyHeatmap(heatmap.Heatmap):

    def __init__(self, vizualization):
        self.update_colors(vizualization)

    def get_all_colors(self):
        """
        Returns all colors in a dict by x by y
        """
        return self.all_colors
    
    def update_colors(self, vizualization):
        """
        Prepares an entropy heatmap

        This method is a heatmap_builder returning a list of patches to be
        plotted somewhere
        The maximum entropy for the Vizualization is calculated and used as
        normalization parameter,
        The plot is actually a logplot as it is more eye-friendly
        ..seealso:: add_heatmap

        """
        self.all_colors = [[0 for _ in range(vizualization.resolution)] for _ in range(vizualization.resolution) ]
        logging.info('heatmap entropy: drawing')
        centroids_label = vizualization.clusterizer.predict(vizualization.mesh_centroids)
        
        entropys = []

        ordered_class_list = list(vizualization.nb_of_individual_by_true_output.keys())
        ordered_class_list.sort()
        global_list = []
        for class_ in ordered_class_list:
            global_list.append(vizualization.nb_of_individual_by_true_output.get(class_))
        global_array = np.array(global_list)
        global_entropy = np.log(global_array / np.sum(global_array))

        for index, (x, y) in enumerate(vizualization.mesh_centroids):

            current_centroid_label = centroids_label[index]

            nb_of_point_by_class      = vizualization.nb_of_points_by_class_by_cluster.get(current_centroid_label, {})
            nb_of_point_by_class_list = [nb_of_point_by_class.get(class_, 0) for class_ in ordered_class_list]
           
            try:
                cluster_is_empty = not bool(vizualization.index_by_cluster_label[current_centroid_label])
            except KeyError:
                cluster_is_empty=True

            if (not nb_of_point_by_class) or cluster_is_empty:
                current_entropy = 0
            else:
                current_entropy = (
                    cross_entropy(
                        global_array,
                        nb_of_point_by_class_list,
                        global_entropy=global_entropy
                        )
                    )

            entropys.append(current_entropy)

        min_entropys = min(entropys)
        max_entropys = max(entropys)
        amplitude_entropys = max_entropys - min_entropys
        # this exception is poorly handleded right now
        if float(amplitude_entropys)==0.:
            amplitude_entropys = 1
        logging.info(
                'heatmap entropy: max cross-entropy={} min={}'.format(
                    max_entropys,
                    min_entropys,)
                )

        for index, (x, y) in enumerate(vizualization.mesh_centroids):
            if index > len(entropys):
                current_entropy = min_entropys
            else:
                current_entropy = entropys[index]

            normalized_entropy = ((current_entropy - min_entropys) / amplitude_entropys)

            x_index, y_index = vizualization.get_coordinates_from_index(index)
            self.all_colors[x_index][y_index] = normalized_entropy

        logging.info('heatmap entropy: done')




import os
import shutil
import logging

import sklearn.datasets as dt
import sklearn.preprocessing as pr
from sklearn.linear_model import LogisticRegression

import numpy as np

from vizuka.config import (
        DATA_PATH,
        MODEL_PATH,
        RAW_NAME,
        INPUT_FILE_BASE_NAME,
        REDUCED_DATA_PATH,
        )

predictor_name = 'predict'
version = '_MNIST_example'
extension = '.npz'

logging.info("Downloading MNIST")
mnist = dt.fetch_mldata('MNIST original')

raw = mnist['data']

logging.info("Preprocessing MNIST")
x = [pr.normalize(x.reshape(1,28**2))[0] for x in raw]
y = [str(int(y)) for y in mnist['target']]


logging.info("Fitting a stupid logistic regression on MNIST")
l = LogisticRegression()
l.fit(x,y)
pred = l.predict(x)


logging.info("Saving all files in vizuka/data/*.npz")
np.savez(
        ''.join([
            os.path.join(DATA_PATH, RAW_NAME),
            version,
            extension,
            ]),
        x=[[image.reshape(28,28)] for image in raw],
        columns=['image'],
        )

np.savez(
        ''.join([
            os.path.join(DATA_PATH, INPUT_FILE_BASE_NAME),
            version,
            extension,
            ]),
        x=x,
        y=y,
        )

np.savez(
        ''.join([
            os.path.join(MODEL_PATH, predictor_name),
            version,
            extension,
            ]),
        pred=pred,
        )

shutil.copy(
        os.path.join(os.path.dirname(__file__), '2Dembedding_50_1000_12000_MNIST_example.npz'),
        REDUCED_DATA_PATH,
        )



import numpy as np
import matplotlib
matplotlib.use('Qt5Agg')  # noqa
from matplotlib import pyplot as plt

from vizuka.cluster_viewer.plotter import Plotter

class Density(Plotter):
    def __call__(self, data, fig, spec):
        """
        Plot the density of your :param: data distribution

        :param: data is a list of the observations
        :param: fig is the Figure on which the axe will be drawn
        :param:spec is the GridSpec for the axe

        :return: the axe with the visualization
        """
        if not data:
            return
        axe = plt.Subplot(fig, spec)

        data = [float(d) for d in data]
        bins = 100 # int(len(data)/10)
        hist, bins = np.histogram(data, bins=bins)
        width = .7 *(bins[1] - bins[0])
        center = (bins[:-1] + bins[1:])/2
        
        axe.set_yscale('linear')
        axe.bar(center, hist, align='center', width=width)
        
        fig.add_subplot(axe)
        return axe

class LogDensity(Plotter):
    def __call__(self, data, fig, spec):
        """
        Plot the log-density of your :param: data distribution

        :param: data is a list of the observations
        :param: fig is the Figure on which the axe will be drawn
        :param:spec is the GridSpec for the axe

        :return: the axe with the visualization
        """
        if not data:
            return
        axe = plt.Subplot(fig, spec)

        data = [float(d) for d in data]
        bins = 100 # int(len(data)/10)
        hist, bins = np.histogram(data, bins=bins)
        width = .7 *(bins[1] - bins[0])
        center = (bins[:-1] + bins[1:])/2
        
        axe.set_yscale('log')
        axe.bar(center, hist, align='center', width=width)
        
        fig.add_subplot(axe)
        return axe




"""
Here are the default parameters used in all the package
There are 1.path 2.filenames 3.learning parameters
"""
import os

#
# ALL DEFAULT PATH
#


def path_builder(base_path):
    return [ os.path.join(base_path, relative) for relative in [
                        'set/',
                        'reduced/',
                        'models/',
                        'graph/',
                        'cache/',
                        'saved_clusters/',
                        ]
        ]

# Build all path from one base
BASE_PATH       = os.path.join(os.path.dirname(__file__), 'data/')
      
(
        DATA_PATH,
        REDUCED_DATA_PATH,
        MODEL_PATH,
        GRAPH_PATH,
        CACHE_PATH,
        SAVED_CLUSTERS_PATH,
    
        ) = path_builder(BASE_PATH)

#
# ALL DEFAULT FILENAME
#

# File containing data to be t-SNEed
INPUT_FILE_BASE_NAME = 'preprocessed_inputs_'
RAW_NAME = 'raw_data_'

# t-SNEDed data
REDUCED_DATA_NAME  = '2Dembedding'

# default RN for predictions
DEFAULT_PREDICTOR = 'predict_'

# A version is a string added to the end of each filename
VERSION = 'MNIST_example'


#
#  LEARNING PARAMETERS
#

# Dimension reduction default parameters :
DEFAULT_PROJECTOR = 'tsne'
PROJECTION_DEFAULT_PARAMS = {
        'tsne': {
            'perplexity'  : 50,
           'learning_rate': 1000,
           'n_iter'       : 12000,
           },
        'pca': {
            'nb_dimension' : 2,
            'min_ratio_variance_explained':-1,
            },
        }



from collections import Counter

import wordcloud
import matplotlib
matplotlib.use('Qt5Agg')  # noqa
from matplotlib import pyplot as plt

from vizuka.cluster_viewer.plotter import Plotter

class Wordcloud(Plotter):

    def __call__(self, data, fig, spec):
        """
        Plots some random images found in the :param data: list given

        :param: data is a list of the observations
        :param: fig is the Figure on which the axe will be drawn
        :param:spec is the GridSpec for the axe

        :return: the axe with the visualization
        """
        if not data:
            return
        axe = plt.Subplot(fig, spec)

        # ok fuck let's be stupid for testing purpose
        data = [str(d) for d in data]
        words_freq = Counter(sum([phrase.split(' ') for phrase in data], []))
        del words_freq['']
        wc = wordcloud.WordCloud()
        
        wc.fit_words(words_freq)
        wc.background_color = 'white'
        wc.scale=10 # for better resolution
        wc.relative_scaling = .5 # for sizing words not only on ranks but also reative freq

        axe.axis('off')
        axe.imshow(wc.to_array())
        fig.add_subplot(axe)
        return axe



import importlib
import pkgutil
import inspect

from vizuka.plugins import cluster_viewer as plotter_plugins
from vizuka.cluster_viewer import (
        image,
        density,
        counter,
        wordcloud,
        )


def list_plotter():
    built_in_plotters = {
            'images':image.RandomImages,
            'density':density.Density,
            'logdensity':density.LogDensity,
            'counter':counter.Counter,
            'wordcloud':wordcloud.Wordcloud,
            }

    extra_plotters = {}
    
    for (module_loader, name, ispkg) in pkgutil.iter_modules(plotter_plugins.__path__):
        plugin = importlib.import_module('.plugins.cluster_viewer.'+name, 'vizuka')
        members = inspect.getmembers(plugin)
        for _, class_ in members:
            if class_ in plotter.Plotter.__subclasses__():
                extra_plotters[name] = class_

    return built_in_plotters, extra_plotters
    

def make_plotter(method):
    """
    Returns an axe with the data plotted as you requested

    :param: data is a list of the observations
    :param: fig is the Figure on which the axe will be drawn
    :param:spec is the GridSpec for the axe

    :return: the axe with the visualization
    """
    
    built_in_plotters, extra_plotters = list_plotter()
    available_plotters = {**built_in_plotters, **extra_plotters}
    
    selected_plotter = available_plotters.get(method, None)

    return selected_plotter()
